// Copyright (c) 2006 dean gaudet <dean@arctic.org>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.

#include <inttypes.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "vtime64.h"
#include "rdtsc.h"

#define N_ITER (10000)

// i want cpu family, model, stepping but i'm too lazy to decode
// all the cpuid bits and special extended family/etc..
static void print_fms(void)
{
  unsigned matched = 0;
  unsigned family;
  unsigned model;
  unsigned stepping;
  char vendor[13];
  char buf[1024];

  FILE *f;
  f = fopen("/proc/cpuinfo", "r");
  if (f == NULL) {
    printf("error opening /proc/cpuinfo: %s\n", strerror(errno));
    return;
  }
  while (fgets(buf, sizeof(buf), f)) {
    if (sscanf(buf, "cpu family      : %u", &family) == 1) {
      matched |= 1;
    }
    else if (sscanf(buf, "model           : %u", &model) == 1) {
      matched |= 2;
    }
    else if (sscanf(buf, "stepping        : %u", &stepping) == 1) {
      matched |= 4;
    }
    else if (sscanf(buf, "vendor_id       : %12s", vendor) == 1) {
      matched |= 8;
    }
    if (matched == 0xf) {
      fclose(f);
      printf("%-12s %02x%02x%02x", vendor, family, model, stepping);
#ifdef __x86_64__
      printf(" 64");
#else
      printf(" 32");
#endif
      return;
    }
  }
  printf("didn't find (one of) cpu family, model, and stepping in /proc/cpuinfo!\n");
}

ulong dead_code_me_harder;

int main(int argc, char **argv)
{
  uint64_t a, b;
  ulong i;

  // in each case i measure the cost of actually calculating the
  // cpu number -- and then accumulate the bogus cpu number into
  // this sum variable so the compiler doesn't get rid of it.
  unsigned sum = 0;

  print_fms();

  // the implementations are deliberately in a different file so
  // that gcc doesn't eliminate the call/ret overhead.

#define bench(name)                                     \
  a = rdtsc();                                          \
  for (i = 0; i < N_ITER; ++i) {                        \
    sum += vtime64_##name(0);                           \
  }                                                     \
  b = rdtsc();                                          \
  printf("  "#name" %5.1f", (b - a)/(double)N_ITER);

  bench(sgdt)
  bench(sidt)
  bench(sldt)
  bench(lsl)

  printf("  w/user restart\n");
#undef bench

  print_fms();
#define bench(name)                                     \
  a = rdtsc();                                          \
  for (i = 0; i < N_ITER; ++i) {                        \
    sum += vtime64_restart_##name(0);                   \
  }                                                     \
  b = rdtsc();                                          \
  printf("  "#name" %5.1f", (b - a)/(double)N_ITER);

  bench(sgdt)
  bench(sidt)
  bench(sldt)
  bench(lsl)

  printf("  w/kernel restart\n");

  dead_code_me_harder = sum;

  exit(0);
}

